from dataclasses import dataclass
from functools import cached_property
from ..hdlc.frame import Info
from struct import pack_into, unpack_from
from typing import ClassVar

MAX_CONTENT_LENGTH = 0x16
FORMAT_IDENTIFIER = 0x81
GROUP_IDENTIFIER = 0x80
MAX_INFO_DEFAULT = 128
WINDOW_DEFAULT = 1
MIN_MAX_INFO_FIELD_LENGTH = 32
MAX_MAX_INFO_FIELD_LENGTH = 2030
MIN_WINDOW_SIZE = 1
MAX_WINDOW_SIZE = 7


@dataclass
class Negotiation(Info):
    MAXIMUM_INFORMATION_FIELD_LENGTH_TRANSMIT: ClassVar[int] = 5
    MAXIMUM_INFORMATION_FIELD_LENGTH_RECEIVE: ClassVar[int] = 6
    WINDOW_SIZE_TRANSMIT: ClassVar[int] = 7
    WINDOW_SIZE_RECEIVE: ClassVar[int] = 8
    max_info_transmit: int = MAX_INFO_DEFAULT
    max_info_receive: int = MAX_INFO_DEFAULT
    window_transmit: int = WINDOW_DEFAULT
    window_receive: int = WINDOW_DEFAULT

    def set_from_UA(self, content: bytes):
        """set from UA info"""
        def get_max_information_field(data: memoryview) -> tuple[int, memoryview]:
            data_len = data[0]
            ret = int.from_bytes(data[1:1 + data_len], "big")
            if MIN_MAX_INFO_FIELD_LENGTH <= ret <= MAX_MAX_INFO_FIELD_LENGTH:
                return ret, memoryview(data[1 + data_len:])
            else:
                raise ValueError(F"got max information field {ret}, expected {MIN_MAX_INFO_FIELD_LENGTH}..{MAX_MAX_INFO_FIELD_LENGTH}")

        def get_window_size(data: memoryview) -> tuple[int, memoryview]:
            data_len = data[0]
            ret = int.from_bytes(data[1:1 + data_len], "big")
            if MIN_WINDOW_SIZE <= ret <= MAX_WINDOW_SIZE:
                return ret, memoryview(data[1 + data_len:])
            else:
                raise ValueError(F"got window_size {ret}, expected {MIN_WINDOW_SIZE}..{MAX_WINDOW_SIZE}")

        try:
            i_rx = i_tx = MAX_INFO_DEFAULT
            w_rx = w_tx = WINDOW_DEFAULT
            if len(content) == 0:
                """skip setting"""
            elif content[0] != FORMAT_IDENTIFIER:
                raise ValueError(F"got {content[0]=}, expected {FORMAT_IDENTIFIER}")
            else:
                if content[1] != GROUP_IDENTIFIER:
                    raise ValueError(F"got {content[1]=}, expected {GROUP_IDENTIFIER}")
                else:
                    if content[2] != len(content[3:]):
                        raise ValueError(F"got {content[2]=}, but content has length={len(content)}")
                    else:
                        data = memoryview(content[3:])
                        while data:
                            match data[0]:
                                case self.MAXIMUM_INFORMATION_FIELD_LENGTH_TRANSMIT:
                                    i_rx, data = get_max_information_field(data[1:])
                                case self.MAXIMUM_INFORMATION_FIELD_LENGTH_RECEIVE:
                                    i_tx, data = get_max_information_field(data[1:])
                                case self.WINDOW_SIZE_TRANSMIT:
                                    w_rx, data = get_window_size(data[1:])
                                case self.WINDOW_SIZE_RECEIVE:
                                    w_tx, data = get_window_size(data[1:])
                                case wrong_tag:
                                    raise ValueError(F"got {wrong_tag=}, expected ")
            self.max_info_receive, self.max_info_transmit, self.window_receive, self.window_transmit = i_rx, i_tx, w_rx, w_tx
            if hasattr(self, "SNRM"):
                del self.SNRM
        except IndexError as e:
            raise ValueError(F"got wrong UA {content.hex(' ')}")

    @cached_property
    def SNRM(self) -> memoryview:
        def small_pack(field: int, data: int):
            nonlocal buf, offset
            pack_into(">BBB", buf, offset,
                      field,
                      1,
                      data)
            offset += 3

        def big_pack(field: int, data: int):
            nonlocal buf, offset
            pack_into(">BBH", buf, offset,
                      field,
                      2,
                      data)
            offset += 4

        def window_pack(field: int, data: int):
            nonlocal buf, offset
            pack_into(">BBL", buf, offset,
                      field,
                      4,
                      data)
            offset += 6

        buf = bytearray(MAX_CONTENT_LENGTH)
        offset: int = 3
        if self.max_info_transmit == MAX_INFO_DEFAULT:
            """not send max_info_transmit"""
        elif self.max_info_transmit < 0x100:  # 1 byte length
            small_pack(self.MAXIMUM_INFORMATION_FIELD_LENGTH_TRANSMIT, self.max_info_transmit)
        else:
            big_pack(self.MAXIMUM_INFORMATION_FIELD_LENGTH_TRANSMIT, self.max_info_transmit)
        if self.max_info_receive == MAX_INFO_DEFAULT:
            """not send max_info_receive"""
        elif self.max_info_receive < 0x100:
            small_pack(self.MAXIMUM_INFORMATION_FIELD_LENGTH_RECEIVE, self.max_info_receive)
        else:
            big_pack(self.MAXIMUM_INFORMATION_FIELD_LENGTH_RECEIVE, self.max_info_receive)
        if self.window_transmit != WINDOW_DEFAULT:
            window_pack(self.WINDOW_SIZE_TRANSMIT, self.window_transmit)
        if self.window_receive != WINDOW_DEFAULT:
            window_pack(self.WINDOW_SIZE_RECEIVE, self.window_transmit)
        if offset == 3:
            offset = 0
        else:
            pack_into(">BBB", buf, 0,
                      FORMAT_IDENTIFIER,
                      GROUP_IDENTIFIER,
                      offset-3)
        return memoryview(buf[:offset])

    @property
    def info(self) -> bytes:
        return bytes(self.SNRM)

    @property
    def content(self) -> bytes:
        return bytes(self.SNRM)

    def __len__(self):
        return len(self.SNRM)

    def __str__(self):
        return F"NEGOTIATION: {self.max_info_transmit}->[info_size]->{self.max_info_receive} {self.window_transmit}->[window]->{self.window_receive}"



import unittest
from src.DLMS_SPODES.version import AppVersion


class TestType(unittest.TestCase):
    def test_encode_length(self):
        a = AppVersion.from_str("1.1.9")
        self.assertEqual(str(a), "1.1.9", "str conversion")
        b = AppVersion(1, 1, 9)
        self.assertEqual(a, b, "patch custom equal")
        self.assertNotEqual(AppVersion(1, 3), AppVersion(1, 3, 0), "different semver")
        variants = [AppVersion(1, 0, 13, "d1"), AppVersion(1, 0, 9)]
        self.assertEqual(b.select_nearest(variants), None, "select left")
        variants = [
            AppVersion(1, 0, 13, "d1"),
            AppVersion(1, 1, 8),
            AppVersion(1, 1, 6)]
        self.assertEqual(b.select_nearest(variants), variants[1], "select left")
        self.assertTrue(AppVersion(1, 3, 0) < AppVersion(1, 4, 0), "check1")
        print(b)

    def test_nearest(self):
        ver = AppVersion.from_str("1.1.3")
        ver2 = ver.select_nearest([AppVersion.from_str("1.1.2"), AppVersion.from_str("1.1.1"), AppVersion.from_str("1.0.1")])
        print(ver2)


from functools import lru_cache
from dataclasses import dataclass
from .types import cdt, cst
from . import pdu_enums as pdu
from .cosem_interface_classes import collection
from . import exceptions as exc


@dataclass
class Settings:
    current_ass_id: int


class Client:
    objects: collection.Collection
    settings: Settings

    @lru_cache(maxsize=1000)
    def is_writable(self, ln: cst.LogicalName, index: int, association_id: int = None) -> bool:
        if not association_id:
            association_id = self.current_association.id
        match self.objects.getASSOCIATION(association_id).object_list.get_attr_access(ln, index):
            case pdu.AttributeAccess.NO_ACCESS | pdu.AttributeAccess.READ_ONLY | pdu.AttributeAccess.AUTHENTICATED_READ_ONLY:
                return False
            case pdu.AttributeAccess.WRITE_ONLY | pdu.AttributeAccess.READ_AND_WRITE:
                return True
            case pdu.AttributeAccess.AUTHENTICATED_WRITE_ONLY | pdu.AttributeAccess.AUTHENTICATED_READ_AND_WRITE:
                if self.settings.cipher.security == Security.AUTHENTICATION:
                    return True
                else:
                    return False
            case err:
                raise exc.ITEApplication(F"unsupport access: {err}")

    @lru_cache(maxsize=1000)
    def is_readable(self, ln: cst.LogicalName, index: int, association_id: int = None) -> bool:
        if not association_id:
            association_id = self.current_association.id
        match self.objects.getASSOCIATION(association_id).object_list.get_attr_access(ln, index):
            case pdu.AttributeAccess.NO_ACCESS | pdu.AttributeAccess.WRITE_ONLY | pdu.AttributeAccess.AUTHENTICATED_WRITE_ONLY:
                return False
            case pdu.AttributeAccess.READ_ONLY | pdu.AttributeAccess.READ_AND_WRITE:
                return True
            case pdu.AttributeAccess.AUTHENTICATED_READ_ONLY | pdu.AttributeAccess.AUTHENTICATED_READ_AND_WRITE:
                if self.settings.cipher.security == Security.AUTHENTICATION:
                    return True
                else:
                    return False
            case err:
                raise exc.ITEApplication(F"unsupport access: {err}")




import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview, ic
from src.DLMS_SPODES.cosem_interface_classes import implementations as impl
from src.DLMS_SPODES.version import AppVersion


class TestType(unittest.TestCase):

    def test_set_image_block_transfer(self):
        col = collection.get_collection(
            manufacturer=b"KPZ",
            server_type=cdt.OctetString("4d324d5f33"),
            server_ver=AppVersion.from_str("1.4.15"))
        obj: collection.ImageTransfer = col.get_object("0.0.44.0.0.255")
        obj.image_block_transfer.set((2, bytearray(b"1234")))
        print(obj)




from __future__ import annotations
from typing import Dict, Type, Union
from abc import ABC, abstractmethod
from struct import pack


_services: Dict[int, Type[Union[AARE, ]]] = dict()
""" COSEM Services dictionary """


class Tag:
    """ APDU tag TODO: """
    value: int

    def __init__(self, value: int):
        self.value = value

    def __get__(self, instance, owner: COSEMPdu) -> int:
        return self.value

    def __set__(self, instance, value):
        raise ValueError("Tag no supported change")


class COSEMPdu(ABC):
    """ TODO: """
    tag: Tag

    def __init__(self, content: bytearray):
        if content.pop(0) != self.tag:
            raise ValueError(F'Wrong {self.__class__.__name__} tag, expected {self.tag}')
        length = content.pop(0)
        if len(content) != length:
            raise ValueError(F'Wrong PDU length, expected {length}, got {len(content)}')

    def __init_subclass__(cls, **kwargs):
        """ register subclass in _types with unique tag  """
        super().__init_subclass__(**kwargs)
        if hasattr(cls, 'tag'):
            _services.setdefault(cls.tag, cls)
        else:
            """ Handler for subclass without tag """

    @classmethod
    def from_content(cls, content: bytearray):
        tag = content[0]
        service = _services.get(tag, None)(content)
        if service is None:
            raise ValueError(F'service with tag:{tag} is absence in Service Dictionary')
        else:
            return service

    @property
    @abstractmethod
    def info(self) -> bytes:
        """ return info. All services, objects etc... TODO: """

    def content(self) -> bytes:
        return pack('BB', self.tag, len(self.info)) + self.info


class ACSE(COSEMPdu, ABC):
    """ TODO: """


class AARE(ACSE):
    """ TODO: """
    tag = Tag(61)

    def __init__(self, content: bytearray):
        super(AARE, self).__init__(content)

    @property
    def info(self) -> bytes:
        return bytes()


class AARQ(ACSE):
    """ TODO: """
    tag = Tag(60)

    def __init__(self, content: bytearray):
        super(AARQ, self).__init__(content)

    @property
    def info(self) -> bytes:
        return bytes()


if __name__ == '__main__':
    a = COSEMPdu.from_content(bytearray((61,3,1,2,4)))
    b = a.content().hex(' ')
    print(a)



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview
from src.DLMS_SPODES.cosem_interface_classes import implementations as impl
from src.DLMS_SPODES.version import AppVersion


class TestType(unittest.TestCase):

    def test_Data(self):
        obj = collection.DemandRegisterVer0("1.0.1.4.0.255")
        self.assertEqual(obj.CLASS_ID, overview.ClassID.DEMAND_REGISTER, "class_ID")
        self.assertEqual(obj.get_attr_length(), 9, "attributes amount")
        obj.set_attr(2, cdt.Unsigned(8).encoding)
        self.assertRaises(ValueError, obj.set_attr, 3, cdt.LongUnsigned(8).encoding, "raising for other type assign")
        self.assertEqual(obj.last_average_value, None, "check for None value")
        print(obj.current_average_value, obj.last_average_value)



from . import ver0
from ..__class_init__ import *
from ...types import choices


class SecurityPolicyVer1(cdt.FlagMixin, cdt.Enum, elements=tuple(range(8))):
    """ Enforces authentication and/or encryption and/or digital signature using the security algorithms available within security suite. It applies independently for requests and
     responses. When enum value is interpreted as an unsigned, the meaning of each bit is as shown below """


class CertificateEntity(cdt.Enum, elements=(0, 1, 2, 3)):
    """TODO:"""


class CertificateType(cdt.Enum, elements=(0, 1, 2, 3)):
    """TODO:"""


class SecuritySuite(ver0.SecuritySuite, elements=(0, 1, 2)):
    """Version 0 extension"""
    AES_GCM_128_AUT_ENCR_ECDSA_P_256_DIG_SIGN_ECDH_P_256_KEY_AGR_SHA_256_HASH_V44_COMPR_AND_AES_128_KEY_WRAP = 1
    AES_GCM_256_AUT_ENCR_ECDSA_P_384_DIG_SIGN_ECDH_P_384_KEY_AGR_SHA_384_HASH_V44_COMPR_AND_AES_256_KEY_WRAP = 2


class CertificateInfo(cdt.Structure):
    """ TODO: """
    certificate_entity: CertificateEntity
    certificate_type: CertificateType
    serial_number: cdt.OctetString
    issuer: cdt.OctetString
    subject: cdt.OctetString
    subject_alt_name: cdt.OctetString


class Certificates(cdt.Array):
    """Carries information on the X.509 v3 Certificates available and stored in the server"""
    TYPE = CertificateInfo


class KeyID(cdt.Enum, elements=(0, 1, 2, 3)):
    """Version 0 extension"""


class KeyTransferData(cdt.Structure):
    """ TODO: """
    key_id: KeyID
    key_wrapped: cdt.OctetString


class KeyTransfer(cdt.Array):
    """ Array of key_transfer_data """
    TYPE = KeyTransferData


class KeyAgreementData(cdt.Structure):
    """ TODO: """
    key_id: KeyID
    key_data: cdt.OctetString


class KeyAgreement(cdt.Array):
    """ Array of key_agreement_data """
    TYPE = KeyAgreementData


class KeyPair(cdt.Enum, elements=(0, 1, 2)):
    """TODO:"""


class CertificateIdentificationByEntity(cdt.Structure):
    """ TODO: """
    certificate_entity: CertificateEntity
    certificate_type: CertificateType
    system_title: cdt.OctetString


class CertificateIdentificationBySerial(cdt.Structure):
    """ TODO: """
    serial_number: cdt.OctetString
    issuer: cdt.OctetString


class CertificateIdentificationType(cdt.Enum, elements=(0, 1)):
    """TODO:"""


class CertificationIdentificationOption(ut.CHOICE):
    TYPE = cdt.Structure
    ELEMENTS = {0: ut.SequenceElement('by entity', CertificateIdentificationByEntity),
                1: ut.SequenceElement('by serial', CertificateIdentificationBySerial)}


certification_identification_option = CertificationIdentificationOption()


class CertificateIdentification(choices.StructureMixin, cdt.Structure):
    """Override several methods of cdt.Structure. It limited Structure."""
    certificate_identification_type: CertificateIdentificationType
    certification_identification_options: certification_identification_option


class SecuritySetup(ver0.SecuritySetup):
    VERSION = Version.V1
    A_ELEMENTS = (ic.ICAElement("security_policy", SecurityPolicyVer1),
                  ic.ICAElement("security_suite", SecuritySuite),
                  ver0.SecuritySetup.get_attr_element(4),
                  ver0.SecuritySetup.get_attr_element(5),
                  ic.ICAElement("certificates", Certificates, classifier=ic.Classifier.DYNAMIC))

    M_ELEMENTS = (ic.ICMElement("security_activate", SecurityPolicyVer1),
                  ic.ICMElement("key_transfer", KeyTransfer),
                  ic.ICMElement("key_agreement", KeyAgreement),
                  ic.ICMElement("generate_key_pair", KeyPair),
                  ic.ICMElement("generate_certificate_request", KeyPair),
                  ic.ICMElement("import_certificate", cdt.OctetString),
                  ic.ICMElement("export_certificate", CertificateIdentification),
                  ic.ICMElement("remove_certificate", CertificateIdentification))

    def characteristics_init(self):
        self.set_attr(2, None)
        self.set_attr(3, None)
        self.set_attr(6, None)

    @property
    def security_policy(self) -> SecurityPolicyVer1:
        return self.get_attr(2)

    @property
    def security_suite(self) -> SecuritySuite:
        return self.get_attr(3)

    @property
    def certificates(self) -> Certificates:
        return self.get_attr(6)

    @property
    def security_activate(self) -> SecurityPolicyVer1:
        return self.get_meth(1)

    @property
    def key_transfer(self) -> KeyTransferData:
        return self .get_meth(2)

    def key_agreement(self) -> KeyAgreement:
        return self .get_meth(3)

    def generate_certificate_request(self) -> KeyPair:
        return self .get_meth(4)

    def import_certificate(self) -> cdt.OctetString:
        return self .get_meth(5)

    def export_certificate(self) -> CertificateIdentification:
        return self .get_meth(6)

    def remove_certificate(self) -> CertificateIdentification:
        return self .get_meth(7)



from ...types import common_data_types as cdt


class LDN(cdt.OctetString):
    """for ldn. todo: check length in initialisation"""
    def manufacturer(self) -> bytes:
        return self.contents[:3]
